cdef extern from "gbinder/gbinder_types.h":
    ctypedef struct GBinderBridge:
        pass

    ctypedef struct GBinderClient:
        pass

    ctypedef struct GBinderIpc:
        pass

    ctypedef struct GBinderLocalObject:
        pass

    ctypedef struct GBinderLocalReply:
        pass

    ctypedef struct GBinderLocalRequest:
        pass

    ctypedef struct GBinderReader:
        pass
    
    ctypedef struct GBinderRemoteObject:
        pass

    ctypedef struct GBinderRemoteReply:
        pass

    ctypedef struct GBinderRemoteRequest:
        pass

    ctypedef struct GBinderServiceName:
        pass

    ctypedef struct GBinderServiceManager:
        pass

    ctypedef struct GBinderWriter:
        pass

    ctypedef enum GBINDER_STABILITY_LEVEL:
        GBINDER_STABILITY_UNDECLARED = 0
        GBINDER_STABILITY_VENDOR = 0x03
        GBINDER_STABILITY_SYSTEM = 0x0c
        GBINDER_STABILITY_VINTF = 0x3f

    ctypedef GBinderLocalReply* (*GBinderLocalTransactFunc)(GBinderLocalObject* obj, GBinderRemoteRequest* req, unsigned int code, unsigned int flags, int* status, void* user_data)

from libc.stdint cimport int64_t, uint64_t

cdef extern from "gbinder/gbinder_servicemanager.h":
    ctypedef void (*GBinderServiceManagerFunc)(GBinderServiceManager* sm, void* user_data)
    ctypedef bint (*GBinderServiceManagerListFunc)(GBinderServiceManager* sm, char** services, void* user_data)
    ctypedef void (*GBinderServiceManagerGetServiceFunc)(GBinderServiceManager* sm, GBinderRemoteObject* obj, int status, void* user_data)
    ctypedef void (*GBinderServiceManagerAddServiceFunc)(GBinderServiceManager* sm, int status, void* user_data)
    ctypedef void (*GBinderServiceManagerRegistrationFunc)(GBinderServiceManager* sm, const char* name, void* user_data)

    GBinderServiceManager* gbinder_servicemanager_new2(const char* dev, const char* sm_protocol, const char* rpc_protocol)
    GBinderServiceManager* gbinder_servicemanager_new(const char* dev)
    GBinderLocalObject* gbinder_servicemanager_new_local_object(GBinderServiceManager* sm, const char* iface, GBinderLocalTransactFunc handler, void* user_data)
    GBinderLocalObject* gbinder_servicemanager_new_local_object2(GBinderServiceManager* sm, const char* const* ifaces, GBinderLocalTransactFunc handler, void* user_data)

    GBinderServiceManager* gbinder_servicemanager_ref(GBinderServiceManager* sm)
    void gbinder_servicemanager_unref(GBinderServiceManager* sm)

    bint gbinder_servicemanager_is_present(GBinderServiceManager* sm)
    bint gbinder_servicemanager_wait(GBinderServiceManager* sm, long max_wait_ms)
    unsigned long gbinder_servicemanager_list(GBinderServiceManager* sm, GBinderServiceManagerListFunc func, void* user_data)
    char** gbinder_servicemanager_list_sync(GBinderServiceManager* sm)
    unsigned long gbinder_servicemanager_get_service(GBinderServiceManager* sm, const char* name, GBinderServiceManagerGetServiceFunc func, void* user_data)
    GBinderRemoteObject* gbinder_servicemanager_get_service_sync(GBinderServiceManager* sm, const char* name, int* status)

    unsigned long gbinder_servicemanager_add_service(GBinderServiceManager* sm, const char* name, GBinderLocalObject* obj, GBinderServiceManagerAddServiceFunc func, void* user_data)
    int gbinder_servicemanager_add_service_sync(GBinderServiceManager* sm, const char* name, GBinderLocalObject* obj)
    void gbinder_servicemanager_cancel(GBinderServiceManager* sm, unsigned long id)
    unsigned long gbinder_servicemanager_add_presence_handler(GBinderServiceManager* sm, GBinderServiceManagerFunc func, void* user_data)
    unsigned long gbinder_servicemanager_add_registration_handler(GBinderServiceManager* sm, const char* name, GBinderServiceManagerRegistrationFunc func, void* user_data)

    void gbinder_servicemanager_remove_handler(GBinderServiceManager* sm, unsigned long id)
    void gbinder_servicemanager_remove_handlers(GBinderServiceManager* sm, unsigned long* ids, unsigned int count)

cdef extern from "gbinder/gbinder_buffer.h":
    ctypedef struct GBinderBuffer:
        void* data
        unsigned long size

    void gbinder_buffer_free(GBinderBuffer* buf)

cdef extern from "gbinder/gbinder_bridge.h":
    GBinderBridge* gbinder_bridge_new(const char* name, const char* const* ifaces, GBinderServiceManager* src, GBinderServiceManager* dest)
    GBinderBridge* gbinder_bridge_new2(const char* src_name, const char* dest_name, const char* const* ifaces, GBinderServiceManager* src, GBinderServiceManager* dest)

    void gbinder_bridge_free(GBinderBridge* bridge)

cdef extern from "gbinder/gbinder_client.h":
    ctypedef struct GBinderClientIfaceInfo:
        const char* iface
        unsigned int last_code

    ctypedef void (*GBinderClientReplyFunc)(GBinderClient* client, GBinderRemoteReply* reply, int status, void* user_data)
    ctypedef void (*GDestroyNotify)(void* data)

    GBinderClient* gbinder_client_new(GBinderRemoteObject* object, const char* iface)
    GBinderClient* gbinder_client_new2(GBinderRemoteObject* object, const GBinderClientIfaceInfo* ifaces, unsigned long count)
    GBinderClient* gbinder_client_ref(GBinderClient* client)
    void gbinder_client_unref(GBinderClient* client)

    const char* gbinder_client_interface(GBinderClient* client)
    const char* gbinder_client_interface2(GBinderClient* client, unsigned int code)

    GBinderLocalRequest* gbinder_client_new_request(GBinderClient* client)
    GBinderLocalRequest* gbinder_client_new_request2(GBinderClient* client, unsigned int code)

    GBinderRemoteReply* gbinder_client_transact_sync_reply(GBinderClient* client, unsigned int code, GBinderLocalRequest* req, int* status)
    int gbinder_client_transact_sync_oneway(GBinderClient* client, unsigned int code, GBinderLocalRequest* req)
    unsigned long gbinder_client_transact(GBinderClient* client, unsigned int code, unsigned int flags, GBinderLocalRequest* req, GBinderClientReplyFunc reply, GDestroyNotify destroy, void* user_data)

    void gbinder_client_cancel(GBinderClient* client, unsigned long id)

cdef extern from "gbinder/gbinder_local_object.h":
    GBinderLocalObject* gbinder_local_object_new(GBinderIpc* ipc, const char* const* ifaces, GBinderLocalTransactFunc handler, void* user_data)

    GBinderLocalObject* gbinder_local_object_ref(GBinderLocalObject* obj)
    void gbinder_local_object_unref(GBinderLocalObject* obj)

    void gbinder_local_object_drop(GBinderLocalObject* obj)
    GBinderLocalReply* gbinder_local_object_new_reply(GBinderLocalObject* obj)
    void gbinder_local_object_set_stability(GBinderLocalObject* obj, GBINDER_STABILITY_LEVEL stability)

cdef extern from "gbinder/gbinder_local_reply.h":
    GBinderLocalReply* gbinder_local_reply_ref(GBinderLocalReply* reply)
    void gbinder_local_reply_unref(GBinderLocalReply* reply)

    void gbinder_local_reply_init_writer(GBinderLocalReply* reply, GBinderWriter* writer)
    void gbinder_local_reply_cleanup(GBinderLocalReply* reply, GDestroyNotify destroy, void* pointer)

    GBinderLocalReply* gbinder_local_reply_append_bool(GBinderLocalReply* reply, bint value)
    GBinderLocalReply* gbinder_local_reply_append_int32(GBinderLocalReply* reply, unsigned int value)
    GBinderLocalReply* gbinder_local_reply_append_int64(GBinderLocalReply* reply, unsigned long value)
    GBinderLocalReply* gbinder_local_reply_append_float(GBinderLocalReply* reply, float value)
    GBinderLocalReply* gbinder_local_reply_append_double(GBinderLocalReply* reply, double value)
    GBinderLocalReply* gbinder_local_reply_append_string8(GBinderLocalReply* reply, const char* str)
    GBinderLocalReply* gbinder_local_reply_append_string16(GBinderLocalReply* reply, const char* utf8)
    GBinderLocalReply* gbinder_local_reply_append_hidl_string(GBinderLocalReply* reply, const char* str)
    GBinderLocalReply* gbinder_local_reply_append_hidl_string_vec(GBinderLocalReply* reply, const char* strv[], signed long count)
    GBinderLocalReply* gbinder_local_reply_append_local_object(GBinderLocalReply* reply, GBinderLocalObject* obj)
    GBinderLocalReply* gbinder_local_reply_append_remote_object(GBinderLocalReply* reply, GBinderRemoteObject* obj)

cdef extern from "gbinder/gbinder_local_request.h":
    GBinderLocalRequest* gbinder_local_request_ref(GBinderLocalRequest* request)
    void gbinder_local_request_unref(GBinderLocalRequest* request)

    void gbinder_local_request_init_writer(GBinderLocalRequest* request, GBinderWriter* writer)
    void gbinder_local_request_cleanup(GBinderLocalRequest* request, GDestroyNotify destroy, void* pointer)

    GBinderLocalRequest* gbinder_local_request_append_bool(GBinderLocalRequest* request, bint value)
    GBinderLocalRequest* gbinder_local_request_append_int32(GBinderLocalRequest* request, unsigned int value)
    GBinderLocalRequest* gbinder_local_request_append_int64(GBinderLocalRequest* request, unsigned long value)
    GBinderLocalRequest* gbinder_local_request_append_float(GBinderLocalRequest* request, float value)
    GBinderLocalRequest* gbinder_local_request_append_double(GBinderLocalRequest* request, double value)
    GBinderLocalRequest* gbinder_local_request_append_string8(GBinderLocalRequest* request, const char* str)
    GBinderLocalRequest* gbinder_local_request_append_string16(GBinderLocalRequest* request, const char* utf8)
    GBinderLocalRequest* gbinder_local_request_append_hidl_string(GBinderLocalRequest* request, const char* str)
    GBinderLocalRequest* gbinder_local_request_append_hidl_string_vec(GBinderLocalRequest* request, const char* strv[], signed long count)
    GBinderLocalRequest* gbinder_local_request_append_local_object(GBinderLocalRequest* request, GBinderLocalObject* obj)
    GBinderLocalRequest* gbinder_local_request_append_remote_object(GBinderLocalRequest* request, GBinderRemoteObject* obj)

cdef extern from "gbinder/gbinder_reader.h":
    bint gbinder_reader_at_end(const GBinderReader* reader)
    bint gbinder_reader_read_byte(GBinderReader* reader, unsigned char* value)
    bint gbinder_reader_read_bool(GBinderReader* reader, bint* value)
    bint gbinder_reader_read_int32(GBinderReader* reader, signed int* value)
    bint gbinder_reader_read_uint32(GBinderReader* reader, unsigned int* value)
    bint gbinder_reader_read_int64(GBinderReader* reader, int64_t* value)
    bint gbinder_reader_read_uint64(GBinderReader* reader, uint64_t* value)
    bint gbinder_reader_read_float(GBinderReader* reader, float* value)
    bint gbinder_reader_read_double(GBinderReader* reader, double* value)
    int gbinder_reader_read_fd(GBinderReader* reader)
    int gbinder_reader_read_dup_fd(GBinderReader* reader)
    bint gbinder_reader_read_nullable_object(GBinderReader* reader, GBinderRemoteObject** obj)

    GBinderRemoteObject* gbinder_reader_read_object(GBinderReader* reader)
    GBinderBuffer* gbinder_reader_read_buffer(GBinderReader* reader)

    const void* gbinder_reader_read_hidl_struct1(GBinderReader* reader, unsigned long size)
    const void* gbinder_reader_read_hidl_vec(GBinderReader* reader, size_t* count, size_t* elemsize)
    const void* gbinder_reader_read_hidl_vec1(GBinderReader* reader, size_t* count, unsigned int expected_elemsize)
    char* gbinder_reader_read_hidl_string(GBinderReader* reader)
    const char* gbinder_reader_read_hidl_string_c(GBinderReader* reader)
    char** gbinder_reader_read_hidl_string_vec(GBinderReader* reader)

    bint gbinder_reader_skip_buffer(GBinderReader* reader)

    const char* gbinder_reader_read_string8(GBinderReader* reader)
    char* gbinder_reader_read_string16(GBinderReader* reader)
    bint gbinder_reader_read_nullable_string16(GBinderReader* reader, char** out)
    #bint gbinder_reader_read_nullable_string16_utf16(GBinderReader* reader, const unsigned short** out, unsigned long* len)
    #const unsigned short* gbinder_reader_read_string16_utf16(GBinderReader* reader,unsigned long* len)
    bint gbinder_reader_skip_string16(GBinderReader* reader)
    const void* gbinder_reader_read_byte_array(GBinderReader* reader, size_t* len)
    unsigned long gbinder_reader_bytes_read(const GBinderReader* reader)
    unsigned long gbinder_reader_bytes_remaining(const GBinderReader* reader)
    #void gbinder_reader_copy(GBinderReader* dest, const GBinderReader* src)

cdef extern from "gbinder/gbinder_remote_object.h":
    ctypedef void (*GBinderRemoteObjectNotifyFunc)(GBinderRemoteObject* obj, void* user_data)

    GBinderRemoteObject* gbinder_remote_object_ref(GBinderRemoteObject* obj)
    void gbinder_remote_object_unref(GBinderRemoteObject* obj)

    GBinderIpc* gbinder_remote_object_ipc(GBinderRemoteObject* obj)
    bint gbinder_remote_object_is_dead(GBinderRemoteObject* obj)
    unsigned long gbinder_remote_object_add_death_handler(GBinderRemoteObject* obj, GBinderRemoteObjectNotifyFunc func, void* user_data)
    void gbinder_remote_object_remove_handler(GBinderRemoteObject* obj, unsigned long id)

cdef extern from "gbinder/gbinder_remote_reply.h":
    GBinderRemoteReply* gbinder_remote_reply_ref(GBinderRemoteReply* reply)
    void gbinder_remote_reply_unref(GBinderRemoteReply* reply)
    void gbinder_remote_reply_init_reader(GBinderRemoteReply* reply, GBinderReader* reader)
    GBinderLocalReply* gbinder_remote_reply_copy_to_local(GBinderRemoteReply* reply)

    bint gbinder_remote_reply_read_int32(GBinderRemoteReply* reply, signed int* value)
    bint gbinder_remote_reply_read_uint32(GBinderRemoteReply* reply, unsigned int* value)
    bint gbinder_remote_reply_read_int64(GBinderRemoteReply* reply, int64_t* value)
    bint gbinder_remote_reply_read_uint64(GBinderRemoteReply* reply, uint64_t* value)
    const char* gbinder_remote_reply_read_string8(GBinderRemoteReply* reply)
    char* gbinder_remote_reply_read_string16(GBinderRemoteReply* reply)
    GBinderRemoteObject* gbinder_remote_reply_read_object(GBinderRemoteReply* reply)

cdef extern from "gbinder/gbinder_remote_request.h":
    GBinderRemoteRequest* gbinder_remote_request_ref(GBinderRemoteRequest* req)
    void gbinder_remote_request_unref(GBinderRemoteRequest* req)
    void gbinder_remote_request_init_reader(GBinderRemoteRequest* req, GBinderReader* reader)
    GBinderLocalRequest* gbinder_remote_request_copy_to_local(GBinderRemoteRequest* req)

    const char* gbinder_remote_request_interface(GBinderRemoteRequest* req)
    int gbinder_remote_request_sender_pid(GBinderRemoteRequest* req)
    short gbinder_remote_request_sender_euid(GBinderRemoteRequest* req)
    void gbinder_remote_request_block(GBinderRemoteRequest* req)
    void gbinder_remote_request_complete(GBinderRemoteRequest* req, GBinderLocalReply* reply, int status)

    bint gbinder_remote_request_read_int32(GBinderRemoteRequest* req, signed int* value)
    bint gbinder_remote_request_read_uint32(GBinderRemoteRequest* req, unsigned int* value)
    bint gbinder_remote_request_read_int64(GBinderRemoteRequest* req, int64_t* value)
    bint gbinder_remote_request_read_uint64(GBinderRemoteRequest* req, uint64_t* value)
    const char* gbinder_remote_request_read_string8(GBinderRemoteRequest* req)
    char* gbinder_remote_request_read_string16(GBinderRemoteRequest* req)
    GBinderRemoteObject* gbinder_remote_request_read_object(GBinderRemoteRequest* self)

cdef extern from "gbinder/gbinder_writer.h":
    ctypedef struct GBinderParent:
        unsigned int index
        unsigned int offset

    void gbinder_writer_append_int32(GBinderWriter* writer, unsigned int value)
    void gbinder_writer_append_int64(GBinderWriter* writer, unsigned long value)
    void gbinder_writer_append_float(GBinderWriter* writer, float value)
    void gbinder_writer_append_double(GBinderWriter* writer, double value)
    void gbinder_writer_append_string16(GBinderWriter* writer, const char* utf8)
    void gbinder_writer_append_string16_len(GBinderWriter* writer, const char* utf8, signed long num_bytes)
    #void gbinder_writer_append_string16_utf16(GBinderWriter* writer, const unsigned short* utf16, signed long length)
    void gbinder_writer_append_string8(GBinderWriter* writer, const char* str)
    void gbinder_writer_append_string8_len(GBinderWriter* writer, const char* str, unsigned long len)
    void gbinder_writer_append_bool(GBinderWriter* writer, bint value)
    void gbinder_writer_append_bytes(GBinderWriter* writer, const void* data, unsigned long size)
    void gbinder_writer_append_fd(GBinderWriter* writer, int fd)

    unsigned long gbinder_writer_bytes_written(GBinderWriter* writer)

    void gbinder_writer_overwrite_int32(GBinderWriter* writer, unsigned long offset, signed int value)

    unsigned int gbinder_writer_append_buffer_object_with_parent(GBinderWriter* writer, const void* buf, unsigned long len, const GBinderParent* parent)
    unsigned int gbinder_writer_append_buffer_object(GBinderWriter* writer, const void* buf, unsigned long len)

    void gbinder_writer_append_hidl_vec(GBinderWriter* writer, const void* base, unsigned int count, unsigned int elemsize)
    void gbinder_writer_append_hidl_string(GBinderWriter* writer, const char* str)
    void gbinder_writer_append_hidl_string_vec(GBinderWriter* writer, const char* data[], signed long count)
    void gbinder_writer_append_local_object(GBinderWriter* writer, GBinderLocalObject* obj)
    void gbinder_writer_append_remote_object(GBinderWriter* writer, GBinderRemoteObject* obj)
    void gbinder_writer_append_byte_array(GBinderWriter* writer, const void* byte_array, signed int len)

    void* gbinder_writer_malloc(GBinderWriter* writer, unsigned long size)
    void* gbinder_writer_malloc0(GBinderWriter* writer, unsigned long size)
    void* gbinder_writer_memdup(GBinderWriter* writer, const void* buf, unsigned long size)
    void gbinder_writer_add_cleanup(GBinderWriter* writer, GDestroyNotify destroy, void* data)



import gbinder
import logging
import time
from gi.repository import GLib


BINDER_DRIVER = "/dev/binder"
INTERFACE = "gbinder.ITest"
SERVICE_NAME = "gbinder_test"

TRANSACTION_sendTest = 1

serviceManager = gbinder.ServiceManager(BINDER_DRIVER)

class ITest:
    def __init__(self, remote):
        self.client = gbinder.Client(remote, INTERFACE)

    def sendTest(self, arg1):
        request = self.client.new_request()
        request.append_string16(arg1)
        reply, status = self.client.transact_sync_reply(
            TRANSACTION_sendTest, request)

        if status:
            logging.error("Sending reply failed")
        else:
            reader = reply.init_reader()
            rep1 = reader.read_string16()
            return rep1

        return None


def add_service(sendTest):
    def response_handler(req, code, flags):
        logging.debug("Received transaction: ", code)
        reader = req.init_reader()
        local_response = response.new_reply()
        if code == TRANSACTION_sendTest:
            arg1 = reader.read_string16()
            ret1 = sendTest(arg1)
            local_response.append_string16(ret1)

        return local_response, 0

    def binder_presence():
        if serviceManager.is_present():
            status = serviceManager.add_service_sync(SERVICE_NAME, response)

            if status:
                logging.error("Failed to add service " + SERVICE_NAME)
                loop.quit()

    response = serviceManager.new_local_object(INTERFACE, response_handler)
    loop = GLib.MainLoop()
    binder_presence()
    status = serviceManager.add_presence_handler(binder_presence)
    if status:
        loop.run()
    else:
        logging.error("Failed to add presence handler: {}".format(status))


def get_service():
    tries = 1000

    remote, status = serviceManager.get_service_sync(SERVICE_NAME)
    while(not remote):
        if tries > 0:
            logging.warning(
                "Failed to get service {}, trying again...".format(SERVICE_NAME))
            time.sleep(1)
            remote, status = serviceManager.get_service_sync(SERVICE_NAME)
            tries = tries - 1
        else:
            return None

    return ITest(remote)



cimport cgbinder
from libc.stdlib cimport malloc, free

def ensure_binary(s):
    if isinstance(s, bytes):
        return s
    if isinstance(s, str):
        return s.encode()
    raise TypeError("not expecting type '%s'" % type(s))

# Stability levels for local objects (since libgbinder 1.1.40)
STABILITY_UNDECLARED = 0x00
STABILITY_VENDOR = 0x03
STABILITY_SYSTEM = 0x0c
STABILITY_VINTF = 0x3f

cdef class Bridge:
    cdef cgbinder.GBinderBridge* _bridge

    def __cinit__(self, src_name, dest_name, ifaces_list, ServiceManager src, ServiceManager dest):
        cdef size_t l = len(ifaces_list)
        cdef const char** ifaces = <const char**>malloc((l + 1) * sizeof(const char*))
        for i in range(l):
            ifaces[i] = ifaces_list[i]
        ifaces[l] = NULL

        if dest_name is None:
            self._bridge = cgbinder.gbinder_bridge_new(ensure_binary(src_name), ifaces, src._sm, dest._sm)
        else:
            self._bridge = cgbinder.gbinder_bridge_new2(ensure_binary(src_name), ensure_binary(dest_name), ifaces, src._sm, dest._sm)

    def __dealloc__(self):
        if self._bridge is not NULL:
            cgbinder.gbinder_bridge_free(self._bridge)

cdef class RemoteObject:
    cdef cgbinder.GBinderRemoteObject* _object
    cdef public object notify_func

    cdef set_c_object(self, cgbinder.GBinderRemoteObject* remote):
        self._object = remote
        if self._object is not NULL:
            cgbinder.gbinder_remote_object_ref(self._object)

    def __dealloc__(self):
        if self._object is not NULL:
            cgbinder.gbinder_remote_object_unref(self._object)

    def ipc(self):
        if self._object is NULL:
            return None
        ipc = Ipc()
        ipc._ipc = cgbinder.gbinder_remote_object_ipc(self._object)
        return ipc

    def is_dead(self):
        if self._object is not NULL:
            return cgbinder.gbinder_remote_object_is_dead(self._object)
        else:
            return True

    def add_death_handler(self, notify_func):
        if self._object is not NULL:
            self.notify_func = notify_func
            return cgbinder.gbinder_remote_object_add_death_handler(self._object, remote_object_local_notify_func, <void*>self)

    def notify_func_callback(self):
        self.notify_func()

    def remove_handler(self, id):
        if self._object is not NULL:
            cgbinder.gbinder_remote_object_remove_handler(self._object, id)

cdef void remote_object_local_notify_func(cgbinder.GBinderRemoteObject* obj, void* user_data) noexcept with gil:
    (<object>user_data).notify_func_callback()

from libc.stdint cimport int64_t, uint64_t, uint8_t

cdef class RemoteReply:
    cdef cgbinder.GBinderRemoteReply* _reply

    cdef set_c_reply(self, cgbinder.GBinderRemoteReply* reply):
        self._reply = reply
        if self._reply is not NULL:
            cgbinder.gbinder_remote_reply_ref(self._reply)

    def __dealloc__(self):
        if self._reply is not NULL:
            cgbinder.gbinder_remote_reply_unref(self._reply)

    def init_reader(self):
        if self._reply is NULL:
            return None
        reader = Reader()
        cgbinder.gbinder_remote_reply_init_reader(self._reply, &reader._reader)
        return reader

    def copy_to_local(self):
        if self._reply is NULL:
            return None
        reply = LocalReply()
        c_reply = cgbinder.gbinder_remote_reply_copy_to_local(self._reply)
        reply.set_c_reply(c_reply)
        return reply

    def read_int32(self):
        if self._reply is NULL:
            return None
        cdef signed int value
        cdef bint status
        status = cgbinder.gbinder_remote_reply_read_int32(self._reply, &value)
        return status, value

    def read_uint32(self):
        if self._reply is NULL:
            return None
        cdef unsigned int value
        cdef bint status
        status = cgbinder.gbinder_remote_reply_read_uint32(self._reply, &value)
        return status, value

    def read_int64(self):
        if self._reply is NULL:
            return None
        cdef int64_t value
        cdef bint status
        status = cgbinder.gbinder_remote_reply_read_int64(self._reply, &value)
        return status, value

    def read_uint64(self):
        if self._reply is NULL:
            return None
        cdef uint64_t value
        cdef bint status
        status = cgbinder.gbinder_remote_reply_read_uint64(self._reply, &value)
        return status, value

    def read_string8(self):
        if self._reply is NULL:
            return None
        return cgbinder.gbinder_remote_reply_read_string8(self._reply).decode()

    def read_string16(self):
        if self._reply is NULL:
            return None
        return cgbinder.gbinder_remote_reply_read_string16(self._reply).decode()

    def read_object(self):
        if self._reply is NULL:
            return None
        cdef cgbinder.GBinderRemoteObject* c_object = cgbinder.gbinder_remote_reply_read_object(self._reply)
        cdef RemoteObject remote = None
        if c_object is not NULL:
            remote = RemoteObject()
            remote.set_c_object(c_object)

        return remote

cdef class Client:
    cdef cgbinder.GBinderClient* _client
    cdef public object reply_func, destroy_notif

    def __cinit__(self, RemoteObject object, ifaces_list):
        cdef size_t l = len(ifaces_list)
        cdef cgbinder.GBinderClientIfaceInfo* ifaces = <cgbinder.GBinderClientIfaceInfo*>malloc((l) * sizeof(cgbinder.GBinderClientIfaceInfo))
        if isinstance(ifaces_list, list):
            for i in range(l):
                iface = ensure_binary(ifaces_list[i][0])
                ifaces[i].iface = iface
                ifaces[i].last_code = ifaces_list[i][1]

            self._client = cgbinder.gbinder_client_new2(object._object, ifaces, l)
        else:
            free(ifaces)
            self._client = cgbinder.gbinder_client_new(object._object, ensure_binary(ifaces_list))

    def __dealloc__(self):
        if self._client is not NULL:
            cgbinder.gbinder_client_unref(self._client)

    def new_request(self, code = None):
        local = LocalRequest()
        if code is None:
            local._req = cgbinder.gbinder_client_new_request(self._client)
        else:
            local._req = cgbinder.gbinder_client_new_request2(self._client, code)

        return local

    def transact_sync_reply(self, unsigned int code, LocalRequest req):
        cdef int status
        reply = RemoteReply()
        reply._reply = cgbinder.gbinder_client_transact_sync_reply(self._client, code, req._req, &status)
        return reply, status

    def transact_sync_oneway(self, unsigned int code, LocalRequest req):
        return cgbinder.gbinder_client_transact_sync_oneway(self._client, code, req._req)

    def transact(self, unsigned int code, unsigned int flags, LocalRequest req, reply_func, destroy_notif):
        self.reply_func = reply_func
        self.destroy_notif = destroy_notif
        return cgbinder.gbinder_client_transact(self._client, code, flags, req._req, client_reply_func, local_destroy_notif, <void*>self)

    def reply_func_callback(self, reply, status):
        if self.reply_func:
            self.reply_func(reply, status)

    def destroy_notif_callback(self):
        if self.destroy_notif:
            self.destroy_notif()

    def cancel(self, unsigned long id):
        return cgbinder.gbinder_client_cancel(self._client, id)

cdef void client_reply_func(cgbinder.GBinderClient* client, cgbinder.GBinderRemoteReply* c_reply, int status, void* user_data) noexcept with gil:
    reply = RemoteReply()
    reply.set_c_reply(c_reply)
    (<object>user_data).reply_func_callback(reply, status)

cdef void local_destroy_notif(void* user_data) noexcept with gil:
    (<object>user_data).destroy_notif_callback()

cdef class LocalRequest:
    cdef cgbinder.GBinderLocalRequest* _req
    cdef public object destroy_notif

    cdef set_c_req(self, cgbinder.GBinderLocalRequest* req):
        self._req = req
        if self._req is not NULL:
            cgbinder.gbinder_local_request_ref(self._req)

    def __dealloc__(self):
        if self._req is not NULL:
            cgbinder.gbinder_local_request_unref(self._req)

    def init_writer(self):
        if self._req is NULL:
            return None
        writer = Writer()
        cgbinder.gbinder_local_request_init_writer(self._req, &writer._writer)
        return writer

    def cleanup(self, destroy_notif):
        if self._req is not NULL:
            self.destroy_notif = destroy_notif
            cgbinder.gbinder_local_request_cleanup(self._req, local_destroy_notif, <void*>self)

    def destroy_notif_callback(self):
        self.destroy_notif()

    def append_bool(self, bint value):
        if self._req is not NULL:
            cgbinder.gbinder_local_request_append_bool(self._req, value)

    def append_int32(self, unsigned int value):
        if self._req is not NULL:
            cgbinder.gbinder_local_request_append_int32(self._req, value)

    def append_int64(self, unsigned long value):
        if self._req is not NULL:
            cgbinder.gbinder_local_request_append_int64(self._req, value)
    
    def append_float(self, float value):
        if self._req is not NULL:
            cgbinder.gbinder_local_request_append_float(self._req, value)
    
    def append_double(self, double value):
        if self._req is not NULL:
            cgbinder.gbinder_local_request_append_double(self._req, value)

    def append_string8(self, value):
        if self._req is not NULL:
            cgbinder.gbinder_local_request_append_string8(self._req, ensure_binary(value))

    def append_string16(self, value):
        if self._req is not NULL:
            cgbinder.gbinder_local_request_append_string16(self._req, ensure_binary(value))

    def append_hidl_string(self, value):
        if self._req is not NULL:
            cgbinder.gbinder_local_request_append_hidl_string(self._req, ensure_binary(value))

    def append_hidl_string_vec(self, values_list):
        cdef signed long count = len(values_list)
        cdef const char** strv = <const char**>malloc(count * sizeof(const char*))
        for i in range(count):
            value = ensure_binary(values_list[i])
            strv[i] = value

        if self._req is not NULL:
            cgbinder.gbinder_local_request_append_hidl_string_vec(self._req, strv, count)
        else:
            free(strv)

    def append_local_object(self, LocalObject obj):
        if self._req is not NULL:
            cgbinder.gbinder_local_request_append_local_object(self._req, obj._object)

    def append_remote_object(self, RemoteObject obj):
        if self._req is not NULL:
            cgbinder.gbinder_local_request_append_remote_object(self._req, obj._object)

cdef class Ipc:
    cdef cgbinder.GBinderIpc* _ipc

cdef class LocalReply:
    cdef cgbinder.GBinderLocalReply* _reply
    cdef public object destroy_notif

    cdef set_c_reply(self, cgbinder.GBinderLocalReply* reply):
        self._reply = reply
        if self._reply is not NULL:
            cgbinder.gbinder_local_reply_ref(self._reply)

    def __dealloc__(self):
        if self._reply is not NULL:
            cgbinder.gbinder_local_reply_unref(self._reply)

    def init_writer(self):
        if self._reply is NULL:
            return None
        writer = Writer()
        cgbinder.gbinder_local_reply_init_writer(self._reply, &writer._writer)
        return writer

    def cleanup(self, destroy_notif):
        if self._reply is not NULL:
            self.destroy_notif = destroy_notif
            cgbinder.gbinder_local_reply_cleanup(self._reply, local_destroy_notif, <void*>self)

    def destroy_notif_callback(self):
        self.destroy_notif()

    def append_bool(self, bint value):
        if self._reply is not NULL:
            cgbinder.gbinder_local_reply_append_bool(self._reply, value)

    def append_int32(self, unsigned int value):
        if self._reply is not NULL:
            cgbinder.gbinder_local_reply_append_int32(self._reply, value)

    def append_int64(self, unsigned long value):
        if self._reply is not NULL:
            cgbinder.gbinder_local_reply_append_int64(self._reply, value)
    
    def append_float(self, float value):
        if self._reply is not NULL:
            cgbinder.gbinder_local_reply_append_float(self._reply, value)
    
    def append_double(self, double value):
        if self._reply is not NULL:
            cgbinder.gbinder_local_reply_append_double(self._reply, value)

    def append_string8(self, value):
        if self._reply is not NULL:
            cgbinder.gbinder_local_reply_append_string8(self._reply, ensure_binary(value))

    def append_string16(self, value):
        if self._reply is not NULL:
            cgbinder.gbinder_local_reply_append_string16(self._reply, ensure_binary(value))

    def append_hidl_string(self, value):
        if self._reply is not NULL:
            cgbinder.gbinder_local_reply_append_hidl_string(self._reply, ensure_binary(value))

    def append_hidl_string_vec(self, values_list):
        cdef signed long count = len(values_list)
        cdef const char** strv = <const char**>malloc(count * sizeof(const char*))
        for i in range(count):
            value = ensure_binary(values_list[i])
            strv[i] = value

        if self._reply is not NULL:
            cgbinder.gbinder_local_reply_append_hidl_string_vec(self._reply, strv, count)
        else:
            free(strv)

    def append_local_object(self, LocalObject obj):
        if self._reply is not NULL:
            cgbinder.gbinder_local_reply_append_local_object(self._reply, obj._object)

    def append_remote_object(self, RemoteObject obj):
        if self._reply is not NULL:
            cgbinder.gbinder_local_reply_append_remote_object(self._reply, obj._object)

cdef class RemoteRequest:
    cdef cgbinder.GBinderRemoteRequest* _req

    cdef set_c_req(self, cgbinder.GBinderRemoteRequest* req):
        self._req = req
        if self._req is not NULL:
            cgbinder.gbinder_remote_request_ref(self._req)

    def __dealloc__(self):
        if self._req is not NULL:
            cgbinder.gbinder_remote_request_unref(self._req)

    def init_reader(self):
        if self._req is NULL:
            return None
        reader = Reader()
        cgbinder.gbinder_remote_request_init_reader(self._req, &reader._reader)
        return reader

    def copy_to_local(self):
        if self._req is NULL:
            return None
        request = LocalRequest()
        request._req = cgbinder.gbinder_remote_request_copy_to_local(self._req)
        return request

    def interface(self):
        if self._req is NULL:
            return None
        return cgbinder.gbinder_remote_request_interface(self._req)

    def sender_pid(self):
        if self._req is NULL:
            return None
        return cgbinder.gbinder_remote_request_sender_pid(self._req)

    def sender_euid(self):
        if self._req is NULL:
            return None
        return cgbinder.gbinder_remote_request_sender_euid(self._req)

    def block(self):
        if self._req is not NULL:
            cgbinder.gbinder_remote_request_block(self._req)

    def complete(self, LocalReply reply, int status):
        if self._req is not NULL:
            cgbinder.gbinder_remote_request_complete(self._req, reply._reply, status)

    def read_int32(self):
        if self._req is NULL:
            return None
        cdef signed int value
        cdef bint status
        status = cgbinder.gbinder_remote_request_read_int32(self._req, &value)
        return status, value

    def read_uint32(self):
        if self._req is NULL:
            return None
        cdef unsigned int value
        cdef bint status
        status = cgbinder.gbinder_remote_request_read_uint32(self._req, &value)
        return status, value

    def read_int64(self):
        if self._req is NULL:
            return None
        cdef int64_t value
        cdef bint status
        status = cgbinder.gbinder_remote_request_read_int64(self._req, &value)
        return status, value

    def read_uint64(self):
        if self._req is NULL:
            return None
        cdef uint64_t value
        cdef bint status
        status = cgbinder.gbinder_remote_request_read_uint64(self._req, &value)
        return status, value

    def read_string8(self):
        if self._req is NULL:
            return None
        return cgbinder.gbinder_remote_request_read_string8(self._req).decode()

    def read_string16(self):
        if self._req is NULL:
            return None
        return cgbinder.gbinder_remote_request_read_string16(self._req).decode()

    def read_object(self):
        if self._req is NULL:
            return None
        cdef cgbinder.GBinderRemoteObject* c_object = cgbinder.gbinder_remote_request_read_object(self._req)
        cdef RemoteObject remote = None
        if c_object is not NULL:
            remote = RemoteObject()
            remote.set_c_object(c_object)

        return remote

cdef class LocalObject:
    cdef cgbinder.GBinderLocalObject* _object
    cdef public object handler

    def __cinit__(self, Ipc ipc, ifaces_list = [], handler = None):
        cdef size_t l = len(ifaces_list)
        cdef const char** ifaces = <const char**>malloc((l + 1) * sizeof(const char*))
        if handler is not None:
            self.handler = handler
        if ipc is not None:
            for i in range(l):
                iface = ensure_binary(ifaces_list[i])
                ifaces[i] = iface
            ifaces[l] = NULL
            self._object = cgbinder.gbinder_local_object_new(ipc._ipc, ifaces, local_transact_callback, <void*>self)

    def __dealloc__(self):
        if self._object is not NULL:
            cgbinder.gbinder_local_object_unref(self._object)

    def callback(self, req, code, flags):
        return self.handler(req, code, flags)

    cdef set_c_object(self, cgbinder.GBinderLocalObject* object):
        self._object = object
        if self._object is not NULL:
            cgbinder.gbinder_local_object_ref(self._object)

    def drop(self):
        if self._object is not NULL:
            cgbinder.gbinder_local_object_drop(self._object)

    def new_reply(self):
        if self._object is not NULL:
            reply = LocalReply()
            c_reply = cgbinder.gbinder_local_object_new_reply(self._object)
            reply.set_c_reply(c_reply)
            return reply

    def set_stability(self, stability):
        if self._object is not NULL:
            cgbinder.gbinder_local_object_set_stability(self._object, <cgbinder.GBINDER_STABILITY_LEVEL>stability)

cdef cgbinder.GBinderLocalReply* local_transact_callback(cgbinder.GBinderLocalObject* obj, cgbinder.GBinderRemoteRequest* c_req, unsigned int code, unsigned int flags, int* status, void* user_data) noexcept with gil:
    req = RemoteRequest()
    req.set_c_req(c_req)
    reply, status_ret = (<object>user_data).callback(req, code, flags)
    cdef int stat = status_ret
    status[0] = stat
    if reply is None or status_ret < 0:
        return NULL
    return (<LocalReply>reply)._reply

cdef class ServiceManager:
    cdef cgbinder.GBinderServiceManager* _sm
    cdef public object func, list_func, get_service_func

    def __cinit__(self, dev, sm_protocol=None, rpc_protocol=None):
        if sm_protocol and rpc_protocol:
            self._sm = cgbinder.gbinder_servicemanager_new2(ensure_binary(dev), ensure_binary(sm_protocol), ensure_binary(rpc_protocol))
        else:
            self._sm = cgbinder.gbinder_servicemanager_new(ensure_binary(dev))

    def __dealloc__(self):
        if self._sm is not NULL:
            cgbinder.gbinder_servicemanager_unref(self._sm)

    def new_local_object(self, ifaces_list, handler):
        if self._sm is NULL:
            return None

        local_object = LocalObject(None, [], handler)
        cdef size_t l = len(ifaces_list)
        cdef const char** ifaces = <const char**>malloc((l + 1) * sizeof(const char*))

        if isinstance(ifaces_list, list):
            for i in range(l):
                iface = ensure_binary(ifaces_list[i])
                ifaces[i] = iface
            ifaces[l] = NULL
            local_object._object = cgbinder.gbinder_servicemanager_new_local_object2(self._sm, ifaces, local_transact_callback, <void*>local_object)
        else:
            free(ifaces)
            local_object._object = cgbinder.gbinder_servicemanager_new_local_object(self._sm, ensure_binary(ifaces_list), local_transact_callback, <void*>local_object)

        return local_object

    def is_present(self):
        if self._sm is NULL:
            return False
        status = cgbinder.gbinder_servicemanager_is_present(self._sm)
        return status

    def wait(self, max_wait_ms):
        if self._sm is NULL:
            return None
        status = cgbinder.gbinder_servicemanager_wait(self._sm, max_wait_ms)
        return status

    def list(self, list_func):
        if self._sm is NULL:
            return None
        self.list_func = list_func
        status = cgbinder.gbinder_servicemanager_list(self._sm, service_manager_list_func, <void*>self)
        return status

    def list_func_callback(self, services_list):
        return self.list_func(services_list)

    def list_sync(self):
        if self._sm is NULL:
            return None
        cdef char** services = cgbinder.gbinder_servicemanager_list_sync(self._sm)
        services_list = []
        if services == NULL:
            return services_list

        i = 0
        while services[i] != NULL:
            services_list.append(services[i].decode())
            i += 1
        return services_list

    def get_service(self, name, get_service_func):
        if self._sm is NULL:
            return None
        self.get_service_func = get_service_func
        cdef int status = cgbinder.gbinder_servicemanager_get_service(self._sm, ensure_binary(name), service_manager_get_service_func, <void*>self)
        return status

    def get_service_func_callback(self, remote, status):
        self.get_service_func(remote, status)

    def get_service_sync(self, name):
        if self._sm is NULL:
            return None, None
        cdef int status
        cdef cgbinder.GBinderRemoteObject* c_object = cgbinder.gbinder_servicemanager_get_service_sync(self._sm, ensure_binary(name), &status)

        cdef RemoteObject remote = None
        if c_object is not NULL:
            remote = RemoteObject()
            remote.set_c_object(c_object)

        return remote, status

    def add_service(self, name, LocalObject obj, add_service_func):
        if self._sm is NULL:
            return None
        self.add_service_func = add_service_func
        cdef int status = cgbinder.gbinder_servicemanager_add_service(self._sm, ensure_binary(name), obj._object, service_manager_add_service_func, <void*>self)
        return status

    def add_service_func_callback(self, status):
        self.add_service_func(status)

    def add_service_sync(self, name, LocalObject obj):
        if self._sm is NULL:
            return None
        cdef int status = cgbinder.gbinder_servicemanager_add_service_sync(self._sm, ensure_binary(name), obj._object)
        return status

    def cancel(self, id):
        cgbinder.gbinder_servicemanager_cancel(self._sm, id)

    def add_presence_handler(self, func):
        if self._sm is NULL:
            return None
        self.func = func
        cdef unsigned long status = cgbinder.gbinder_servicemanager_add_presence_handler(self._sm, service_manager_func, <void*>self)
        return status

    def func_callback(self):
        self.func()

    def add_registration_handler(self, name, registration_func):
        if self._sm is NULL:
            return None
        self.registration_func = registration_func
        cdef unsigned long status = cgbinder.gbinder_servicemanager_add_registration_handler(self._sm, ensure_binary(name), service_manager_registration_func, <void*>self)
        return status

    def registration_func_callback(self, name):
        self.registration_func(ensure_binary(name))

    def remove_handler(self, id):
        if self._sm is not NULL:
            cgbinder.gbinder_servicemanager_remove_handler(self._sm, id)

    def remove_handlers(self, ids_list):
        cdef unsigned int count = len(ids_list)
        cdef unsigned long* ids = <unsigned long*>malloc((count) * sizeof(unsigned long))
        for i in range(count):
            ids[i] = ids_list[i]

        if self._sm is not NULL:
            cgbinder.gbinder_servicemanager_remove_handlers(self._sm, ids, count)
        else:
            free(ids)

cdef void service_manager_get_service_func(cgbinder.GBinderServiceManager* sm, cgbinder.GBinderRemoteObject* c_object, int status, void* user_data) noexcept with gil:
    remote = RemoteObject()
    remote.set_c_object(c_object)
    (<object>user_data).get_service_func_callback(remote, status)

cdef bint service_manager_list_func(cgbinder.GBinderServiceManager* sm, char** services, void* user_data) noexcept with gil:
    services_list = []
    if services == NULL:
        return services_list

    i = 0
    while services[i] != NULL:
        services_list.append(services[i].decode())
        i += 1
    return (<object>user_data).list_func_callback(services_list)

cdef void service_manager_add_service_func(cgbinder.GBinderServiceManager* sm, int status, void* user_data) noexcept with gil:
    (<object>user_data).add_service_func_callback(status)

cdef void service_manager_func(cgbinder.GBinderServiceManager* sm, void* user_data) noexcept with gil:
    (<object>user_data).func_callback()

cdef void service_manager_registration_func(cgbinder.GBinderServiceManager* sm, const char* name, void* user_data) noexcept with gil:
    (<object>user_data).registration_func_callback(name)

cdef class Buffer:
    cdef cgbinder.GBinderBuffer* _buffer

    cdef set_c_buffer(self, cgbinder.GBinderBuffer* buff):
        self._buffer = buff

    def __dealloc__(self):
        if self._buffer is not NULL:
            cgbinder.gbinder_buffer_free(self._buffer)

    @property
    def data(self):
        if self._buffer is not NULL:
            return <uint8_t[:self._buffer.size]>self._buffer.data

cdef class Writer:
    cdef cgbinder.GBinderWriter _writer
    cdef public object destroy_notif
    # Keep references to data passed to C functions where needed
    cdef list _referenced_objects

    def __init__(self):
        self._referenced_objects = []

    def append_int32(self, unsigned int value):
        cgbinder.gbinder_writer_append_int32(&self._writer, value)

    def append_int64(self, unsigned long value):
        cgbinder.gbinder_writer_append_int64(&self._writer, value)
    
    def append_float(self, float value):
        cgbinder.gbinder_writer_append_float(&self._writer, value)
    
    def append_double(self, double value):
        cgbinder.gbinder_writer_append_double(&self._writer, value)

    def append_string16(self, value):
        cgbinder.gbinder_writer_append_string16(&self._writer, ensure_binary(value))

    def append_string16_len(self, value, signed long num_bytes):
        cgbinder.gbinder_writer_append_string16_len(&self._writer, ensure_binary(value), num_bytes)

    def append_string8(self, value):
        cgbinder.gbinder_writer_append_string8(&self._writer, ensure_binary(value))

    def append_string8_len(self, value, unsigned long len):
        cgbinder.gbinder_writer_append_string8_len(&self._writer, ensure_binary(value), len)

    def append_bool(self, bint value):
        cgbinder.gbinder_writer_append_bool(&self._writer, value)

    def append_bytes(self, value, unsigned long size):
        cgbinder.gbinder_writer_append_bytes(&self._writer, <const uint8_t*>value, size)

    def append_fd(self, int fd):
        cgbinder.gbinder_writer_append_fd(&self._writer, fd)

    def bytes_written(self):
        return cgbinder.gbinder_writer_bytes_written(&self._writer)

    def overwrite_int32(self, unsigned long offset, signed int value):
        cgbinder.gbinder_writer_overwrite_int32(&self._writer, offset, value)

    def append_buffer_object_with_parent(self, buf, parent_tuple):
        cdef cgbinder.GBinderParent* parent = NULL
        parent.index = parent_tuple[0]
        parent.offset = parent_tuple[1]

        # Store a reference to the bytes object to keep it alive
        self._referenced_objects.append(buf)
        cdef const void* ptr = <const void*>(<uint8_t*>buf)

        return cgbinder.gbinder_writer_append_buffer_object_with_parent(&self._writer, ptr, len(buf), parent)

    def append_buffer_object(self, buf):
        # Store a reference to the bytes object to keep it alive
        self._referenced_objects.append(buf)
        cdef const void* ptr = <const void*>(<const uint8_t*>buf)
        return cgbinder.gbinder_writer_append_buffer_object(&self._writer, ptr, len(buf))

    def append_hidl_vec(self, base, unsigned int count=0, unsigned int elemsize=0):
        cdef const void* ptr = NULL

        # Handle null or empty vector
        if base is None or len(base) == 0:
            # Pass NULL pointer with count=0 for empty vectors
            ptr = NULL
            count = 0
        # Accept bytes-like object
        else:
            if count == 0 or elemsize == 0:
                raise ValueError("Count and element size must be provided")
            ptr = <const void*>(<const uint8_t*>base)

        # libgbinder copies the data in gbinder_writer_data_append_hidl_vec,
        # no need to keep the reference, but the caller has to ensure any
        # nested object pointers remain valid

        cgbinder.gbinder_writer_append_hidl_vec(&self._writer, ptr, count, elemsize)

    def append_hidl_string(self, value):
        # Convert to bytes if needed and keep a reference
        binary = ensure_binary(value)
        self._referenced_objects.append(binary)
        cgbinder.gbinder_writer_append_hidl_string(&self._writer, binary)

    def append_hidl_string_vec(self, values_list):
        cdef signed long count = len(values_list)
        cdef const char** strv = <const char**>malloc(count * sizeof(const char*))

        # Keep references to all the binary strings
        binary_strings = []
        for i in range(count):
            binary = ensure_binary(values_list[i])
            binary_strings.append(binary)
            strv[i] = binary

        # Keep a reference to the list of binary strings
        self._referenced_objects.append(binary_strings)

        cgbinder.gbinder_writer_append_hidl_string_vec(&self._writer, strv, count)

    def append_local_object(self, LocalObject obj):
        cgbinder.gbinder_writer_append_local_object(&self._writer, obj._object)

    def append_remote_object(self, RemoteObject obj):
        cgbinder.gbinder_writer_append_remote_object(&self._writer, obj._object)

    def append_byte_array(self, byte_array, signed int len):
        # Keep a reference to the byte array
        self._referenced_objects.append(byte_array)
        cgbinder.gbinder_writer_append_byte_array(&self._writer, <const void*>byte_array, len)

    def malloc(self, byte_array, unsigned long size):
        cdef void* alloc_buf = cgbinder.gbinder_writer_malloc(&self._writer, size)
        return <uint8_t[:size]>alloc_buf

    def malloc0(self, byte_array, unsigned long size):
        cdef void* alloc_buf = cgbinder.gbinder_writer_malloc0(&self._writer, size)
        return <uint8_t[:size]>alloc_buf

    def memdup(self, buf, unsigned long size):
        cdef const void* ptr = <const void*>(<const uint8_t*>buf);
        cdef void* dup_buf = cgbinder.gbinder_writer_memdup(&self._writer, <const void*>buf, size)
        return <uint8_t[:size]>dup_buf

    def add_cleanup(self, destroy_notif):
        self.destroy_notif = destroy_notif
        cgbinder.gbinder_writer_add_cleanup(&self._writer, local_destroy_notif, <void*>self)

    def destroy_notif_callback(self):
        self.destroy_notif()

cdef class Reader:
    cdef cgbinder.GBinderReader _reader

    def at_end(self):
        cgbinder.gbinder_reader_at_end(&self._reader)

    def read_byte(self):
        cdef unsigned char value
        cdef bint status
        status = cgbinder.gbinder_reader_read_byte(&self._reader, &value)
        return status, value

    def read_bool(self):
        cdef bint value
        cdef bint status
        status = cgbinder.gbinder_reader_read_bool(&self._reader, &value)
        return status, value

    def read_int32(self):
        cdef signed int value
        cdef bint status
        status = cgbinder.gbinder_reader_read_int32(&self._reader, &value)
        return status, value

    def read_uint32(self):
        cdef unsigned int value
        cdef bint status
        status = cgbinder.gbinder_reader_read_uint32(&self._reader, &value)
        return status, value

    def read_int64(self):
        cdef int64_t value
        cdef bint status
        status = cgbinder.gbinder_reader_read_int64(&self._reader, &value)
        return status, value

    def read_uint64(self):
        cdef uint64_t value
        cdef bint status
        status = cgbinder.gbinder_reader_read_uint64(&self._reader, &value)
        return status, value

    def read_float(self):
        cdef float value
        cdef bint status
        status = cgbinder.gbinder_reader_read_float(&self._reader, &value)
        return status, value

    def read_double(self):
        cdef double value
        cdef bint status
        status = cgbinder.gbinder_reader_read_double(&self._reader, &value)
        return status, value

    def read_fd(self):
        return cgbinder.gbinder_reader_read_fd(&self._reader)

    def read_dup_fd(self):
        return cgbinder.gbinder_reader_read_dup_fd(&self._reader)

    def read_nullable_object(self):
        cdef cgbinder.GBinderRemoteObject* obj
        cdef bint status
        status = cgbinder.gbinder_reader_read_nullable_object(&self._reader, &obj)
        remote = RemoteObject()
        remote.set_c_object(obj)
        return status, remote

    def read_object(self):
        cdef cgbinder.GBinderRemoteObject* obj = cgbinder.gbinder_reader_read_object(&self._reader)
        remote = RemoteObject()
        remote.set_c_object(obj)
        return remote

    def read_buffer(self):
        cdef cgbinder.GBinderBuffer* value = cgbinder.gbinder_reader_read_buffer(&self._reader)
        buff = Buffer()
        buff.set_c_buffer(value)
        return buff

    def read_hidl_struct1(self, unsigned long size):
        cdef const void* ptr = cgbinder.gbinder_reader_read_hidl_struct1(&self._reader, size)
        if ptr == NULL:
            return None
        return <const uint8_t[:size]>ptr

    def read_hidl_vec(self):
        cdef size_t count, elemsize
        cdef const void* ptr = cgbinder.gbinder_reader_read_hidl_vec(&self._reader, &count, &elemsize)
        if ptr == NULL:
            return None
        return <const uint8_t[:count, :elemsize]>ptr

    def read_hidl_vec1(self, unsigned int expected_elemsize):
        cdef size_t count
        cdef const void* ptr = cgbinder.gbinder_reader_read_hidl_vec1(&self._reader, &count, expected_elemsize)
        if ptr == NULL:
            return None
        return <const uint8_t[:count, :expected_elemsize]>ptr

    def read_hidl_string(self):
        return cgbinder.gbinder_reader_read_hidl_string_c(&self._reader).decode()

    def read_hidl_string_vec(self):
        cdef char** value = cgbinder.gbinder_reader_read_hidl_string_vec(&self._reader)
        string_list = []
        i = 0

        while value[i] != NULL:
            string_list.append(value[i].decode())
            i += 1
        return string_list

    def skip_buffer(self):
        return cgbinder.gbinder_reader_skip_buffer(&self._reader)

    def read_string8(self):
        return cgbinder.gbinder_reader_read_string8(&self._reader).decode()

    def read_string16(self):
        return cgbinder.gbinder_reader_read_string16(&self._reader).decode()

    def nullable_string16(self):
        cdef char* value
        cdef bint status
        status = cgbinder.gbinder_reader_read_nullable_string16(&self._reader, &value)
        return status, value.decode()

    def skip_string16(self):
        return cgbinder.gbinder_reader_skip_string16(&self._reader)

    def read_byte_array(self):
        cdef size_t len
        cdef const void* ptr = cgbinder.gbinder_reader_read_byte_array(&self._reader, &len)
        if ptr == NULL:
            return None
        return <const uint8_t[:len]>ptr

    def bytes_read(self):
        return cgbinder.gbinder_reader_bytes_read(&self._reader)

    def bytes_remaining(self):
        return cgbinder.gbinder_reader_bytes_remaining(&self._reader)



import sys, subprocess
from setuptools import setup, Extension
from Cython.Build import cythonize


def pkgconfig(package, kw):
    flag_map = {'-I': 'include_dirs', '-L': 'library_dirs', '-l': 'libraries'}
    output = subprocess.getoutput(
        'pkg-config --cflags --libs {}'.format(package))
    for token in output.strip().split():
        kw.setdefault(flag_map.get(token[:2]), []).append(token[2:])
    return kw

extension_kwargs = { 'sources': ["gbinder.pyx"] }
extension_kwargs = pkgconfig('libgbinder', extension_kwargs)
if None in extension_kwargs:
    del extension_kwargs[None]
extensions = [Extension('gbinder', **extension_kwargs)]
extensions = cythonize(extensions, compiler_directives={
                       'language_level': "3"})

setup(
    name="gbinder-python",
    description="""Cython extension module for C++ gbinder functions""",
    version="1.3.1",
    author="Erfan Abdi",
    author_email="erfangplus@gmail.com",
    url="https://github.com/erfanoabdi/gbinder-python",
    license="GPL3",
    ext_modules=extensions
)
